/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _FAST_LANDMARK_MAP_HPP
#define _FAST_LANDMARK_MAP_HPP

#include <stdint.h>
#include <vector>

// balanced binary tree map representation that enables fast
// resampling (no copying) in exchange for log(n) insertion and
// modification.  the basic idea comes from log(n) fastslam (e.g.,
// montemerlo's thesis) although my implementation is somewhat
// different.  specifically, every node in the tree stores a landmark,
// as opposed to storing landmarks only at leaves.  landmark "ids" are
// assigned in such a way that sequential insertion leads to a
// balanced tree.  there is no rebalancing since it adds a lot of
// complication to the problem of reusing trees for multiple
// particles.

// a quick overview of binary tree operations is available at:
// http://www.cs.niu.edu/~gbrown/tree.htm

// the same ideas can be applied to kd-trees, although with a lot more
// bookkeeping effort because of the dimension-dependence of the tree
// levels.  some random kd-tree code can be found at:
// http://www.cs.fiu.edu/~weiss/dsaa_c++/Code/KdTree.cpp
// http://www2.imm.dtu.dk/~jab/software.html


//#define DEBUG

// defined in slam.hpp:
class pose_t;
uint32_t max_uint32();
FILE * open_file(const char *prefix, const char *extension, const char *mode);


template <class F> // F == feature type
class map_t
{
public:
  struct bbt_node
  {
    bbt_node *left, *right;
    uint32_t id;    // "node id"
    uint32_t count; // reference count
    F landmark;
  };

protected:
  bbt_node *root;
  uint32_t mysize;

public:
  // okay, not really an "iterator" in the usual sense but it's easier
  // this way
  typedef bbt_node * iterator;
  typedef const bbt_node * const_iterator;

  map_t() : root(0), mysize(0), lcount(1), level(2) {}
  map_t(const map_t &m)
    : root(m.root), mysize(m.mysize), lcount(m.lcount), level(m.level)
  {
    if(root) root->count++;
  }
  ~map_t() { internal_destroy(root); }
  void clear() { internal_destroy(root); }
  map_t & operator=(const map_t &m);
  inline iterator insert(const F &feature) { return insert(feature, genid()); }
  iterator insert(const F &feature, uint32_t id);
  void erase(iterator &f);
  iterator find_id(uint32_t id) const { return internal_find_id(root, id); }
  iterator modify(iterator &f);
  void in_range(const pose_t &p, double range, std::vector<iterator> &out) const;
  uint32_t size() const { return mysize; }
  inline void render() const { internal_render(root); }
  void write_to_file(const char *prefix) const;
  void read_from_file(const char *file);

  template <class Container>
  void get_all_nodes(Container &v) { internal_get_all_nodes(v, root); }

#if 0
  // FIXME: remove
  void print(bbt_node *n = (bbt_node *)0xffffffff) {
    if(n == (bbt_node *)0xffffffff) n = root;
    if(!n)
      return;
    printf("%d/%d: (%f %f) [%d %d]\n", n->id, n->count, n->landmark.x, n->landmark.y,
	   n->left ? n->left->id : -1, n->right ? n->right->id : -1);
    print(n->left);
    print(n->right);
  }
#endif

protected:
  uint32_t lcount, level;
  uint32_t genid();

  void internal_destroy(bbt_node *n);
  bbt_node * internal_insert(bbt_node *n, bbt_node **cur);
  bbt_node * internal_find_id(bbt_node *n, uint32_t id) const;
  bbt_node * internal_modify(bbt_node **f, bbt_node *cur);
  void internal_in_range
    (const pose_t &p, double sq_range,
     std::vector<iterator> &out, bbt_node *n) const;
  void internal_render(bbt_node *n) const;
  void internal_write_to_file(FILE *file, bbt_node *n) const;

  template <class Container>
  void internal_get_all_nodes(Container &v, bbt_node *n)
  {
    if(!n)
      return;
    v.push_back(n);
    internal_get_all_nodes(v, n->left);
    internal_get_all_nodes(v, n->right);
  }
};


//////////////////////////////////////////////////////////////////////
// implementation details
//////////////////////////////////////////////////////////////////////

template <class F>
map_t<F> & map_t<F>::operator=(const map_t<F> &m)
{
  internal_destroy(root);
  root = m.root;
  if(root) root->count++;
  mysize = m.mysize;
  lcount = m.lcount;
  level = m.level;
  return *this;
}

template <class F>
typename map_t<F>::iterator map_t<F>::insert(const F &feature, uint32_t id)
{
  assert(id != 0); // see rbpf.hpp
  bbt_node *n = new bbt_node;
  n->landmark = feature;
  n->left = n->right = 0;
  n->count = 1;
  n->id = id;
  root = internal_insert(n, &root);
  ++mysize;
  return n;
}

template <class F>
void map_t<F>::erase(map_t<F>::iterator &f)
{
  // first, find f in the tree so we know its parent
  bbt_node *parent = 0, *n = root;
  while(n->id != f->id) {
    parent = n;
    if(f->id < n->id)
      n = n->left;
    else
      n = n->right;
  }

  //  printf("n = 0x%x, parent = 0x%x, n->left = 0x%x, n->right = 0x%x, n->right->count = %d, n->count = %d\n",
  //	 n, parent, n->left, n->right, n->right ? n->right->count : -1, n->count);

  if(n->left == 0 && n->right == 0) {
    // no children
    parent = modify(parent);
    if(parent) { // in case n == root
      if(n == parent->left) parent->left = 0;
      else parent->right = 0;
    }
  } else if(n->left == 0 && n->right !=0) {
    // only right child
    parent = modify(parent);
    if(parent) {
      if(n == parent->left) parent->left = n->right;
      else parent->right = n->right;
    } else
      root = n->right;
    n->right->count++;
  } else if(n->left != 0 && n->right == 0) {
    // only left child
    parent = modify(parent);
    if(parent) {
      if(n == parent->left) parent->left = n->left;
      else parent->right = n->left;
    } else
      root = n->left;
    n->left->count++;
  } else if(n->left != 0 && n->left->right == 0) {
    // have left child and it has no right child
    // FIXME: broken case
    return;
    n->left = modify(n->left);
    // FIXME: re-get n, parent of n ???
    n->left->right = n->right;
    if(n->right) n->right->count++;
    if(parent) {
      if(n == parent->left) parent->left = n->left;
      else parent->right = n->left;
    } else
      root = n->left;
    n->left->count++;
  } else { // n->left != 0 && n->left->right != 0
    // have left child and it has right child
    // FIXME: broken case
    return;
    bbt_node *qp = n->left;
    bbt_node *q = qp->right;
    while(q->right != 0) { // find rightmost node in this subtree
      qp = q;
      q = q->right;
    }
    // replace n with the rightmost node
    q = modify(q);
    qp = modify(qp);
    q->right = n->right;
    if(n->right) n->right->count++;
    qp->right = q->left;
    q->left = n->left;
    n->left->count++;
    if(parent) {
      if(n == parent->left) parent->left = q;
      else parent->right = q;
    } else
      root = q;
  }

  if(n->count == 0)
    internal_destroy(n);
  --mysize;
}

template <class F>
typename map_t<F>::iterator map_t<F>::modify(map_t<F>::iterator &f)
{
  if(f == 0)
    return 0;
  if(f->count == 1)
    return f;
  bbt_node *nn = f;
  root = internal_modify(&nn, root);
  return nn;
}

template <class F>
void map_t<F>::in_range
  (const pose_t &p, double range,
   std::vector<map_t<F>::iterator> &out) const
{
  out.clear();
  internal_in_range(p, range*range, out, root);
}

// generate "node ids" that lead to a well-balanced binary tree if we
// use them in a sequence of insertions
template <class F>
uint32_t map_t<F>::genid()
{
  static uint32_t r;
  // start with level = 2 and lcount = 1
  if(lcount > level) {
    level *= 2;
    lcount = 1;
  }
  r = (max_uint32() / level) * lcount;
  lcount += 2;
  return r;
}

template <class F>
void map_t<F>::internal_destroy(bbt_node *n)
{
  if(n == 0)
    return;
  if(n->count <= 1) {
    internal_destroy(n->left);
    internal_destroy(n->right);
    //    printf("deleting 0x%x, count == %d\n", n, n->count);
    delete n;
    n = 0;
  } else
    n->count--;
}

template <class F>
typename map_t<F>::bbt_node *
map_t<F>::internal_insert(bbt_node *n, bbt_node **cur)
{
  if(*cur == 0) {
    *cur = n;
    return n;
  }

  bbt_node *nn = *cur;
  if((*cur)->count > 1) {
    nn = new bbt_node(**cur);
    nn->count = 1;
    (*cur)->count--;
    if(nn->left) nn->left->count++;
    if(nn->right) nn->right->count++;
  }

  if(n->id < nn->id)
    nn->left = internal_insert(n, &(nn->left));
  else
    nn->right = internal_insert(n, &(nn->right));

  return nn;
}

template <class F>
typename map_t<F>::bbt_node *
map_t<F>::internal_find_id(bbt_node *n, uint32_t id) const
{
  if(n == 0)
    return 0;
  if(n->id == id)
    return n;
  if(id < n->id)
    return internal_find_id(n->left, id);
  else
    return internal_find_id(n->right, id);
}

template <class F>
typename map_t<F>::bbt_node *
map_t<F>::internal_modify(bbt_node **f, bbt_node *cur)
{
  assert(cur != 0);

  if(cur->count == 1)
    return cur; // no need to make extra copies from here down

  bbt_node *nn = new bbt_node(*cur);
  nn->count = 1;
  cur->count--;
  if(nn->left) nn->left->count++;
  if(nn->right) nn->right->count++;

  if(cur == *f)
    *f = nn;
  else if((*f)->id < nn->id)
    nn->left = internal_modify(f, nn->left);
  else
    nn->right = internal_modify(f, nn->right);

  return nn;
}

// O(n) since this isn't a kd-tree
template <class F>
void map_t<F>::internal_in_range
  (const pose_t &p, double sq_range, std::vector<map_t<F>::iterator> &out,
   bbt_node *n) const
{
  if(!n)
    return;
  if(squared_distance(p, n->landmark) <= sq_range)
    out.push_back(n);
  internal_in_range(p, sq_range, out, n->left);
  internal_in_range(p, sq_range, out, n->right);
}

template <class F>
void map_t<F>::internal_render(bbt_node *n) const
{
  if(!n)
    return;
  n->landmark.render();
  internal_render(n->left);
  internal_render(n->right);
}

template <class F>
void map_t<F>::write_to_file(const char *prefix) const
{
  FILE *file = open_file(prefix, "lmk", "w");
  if(file) {
    internal_write_to_file(file, root);
    fclose(file);
  }
}

template <class F>
void map_t<F>::read_from_file(const char *file)
{
  clear();
  FILE *f = fopen(file, "r");
  if(f) {
    F feat;
    while(!feof(f)) {
      feat.read_from_file(f);
      insert(feat);
    }
  }
}

template <class F>
void map_t<F>::internal_write_to_file(FILE *file, bbt_node *n) const
{
  if(!n)
    return;
  n->landmark.write_to_file(file);
  internal_write_to_file(file, n->left);
  internal_write_to_file(file, n->right);
}


#if 0
void test_fast_landmark_map()
{
  map_t<point_t> m;
  m.insert(point_t(0,0));
  m.insert(point_t(1,0));
  m.insert(point_t(2,1));
  m.insert(point_t(3,1));
  m.insert(point_t(-3,-2));

  std::vector<map_t<point_t>::iterator> out;
  m.in_range(pose_t(0,0,0), 1, out);
  for(uint32_t i = 0; i < out.size(); ++i)
    printf("%f %f\n", out[i]->landmark.x, out[i]->landmark.y);
  printf("\n");

  map_t<point_t> m2(m);

  map_t<point_t>::iterator n = m.modify(out[0]);
  n->landmark.x = 1;

  printf("m\n"); m.print();
  printf("m2\n"); m2.print();

  m.insert(point_t(8,9));
  printf("m\n"); m.print();
  printf("m2\n"); m2.print();

  m.erase(out[0]);
  printf("m\n"); m.print();
  m.erase(out[1]);
  printf("m\n"); m.print();
}
#endif


#endif // _FAST_LANDMARK_MAP_HPP
