/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _LOGFILE_HPP
#define _LOGFILE_HPP

#include "slam.hpp"
#include <zlib.h>

class logfile_t
{
public:

  // types of data we can handle
  typedef enum {
    ERROR,
    COMMENT,
    ODOMETRY,
    SCAN,
    GPS,
    TIME
  } datatype_t;


  logfile_t()
    : line_count(0), last_time(0), cur_time(0), max_range(0), file(0), fn(0) {}

  virtual ~logfile_t();

  bool open(const char *filename);
  bool reopen() { return open(fn); }
  bool eof() const;

  // read lines from the file until a new scan is read
  bool read_scan();

  // accessors
  pose_t & pose() { return cur_pose; }
  const pose_t & pose() const { return cur_pose; }
  scan_t & scan() { return cur_scan; }
  const scan_t & scan() const { return cur_scan; }
  point_t & gps() { return cur_gps; }
  const point_t & gps() const { return cur_gps; }
  uint32_t line_number() const { return line_count; }
  double time() const { return cur_time; }
  double range_max() const { return max_range; }

protected:

  pose_t cur_pose;           // current laser pose based on odometry estimate
  scan_t cur_scan;           // current laser data
  point_t cur_gps;           // current gps pose

  uint32_t line_count;
  double last_time, cur_time;
  double max_range;

  gzFile file;
  char line[4096];         // file line buffer
  char *fn;
  std::vector<char *> tokens;

  // for use by subclasses to get another line from the log file
  char * read_line();

public:

  // these must be implemented by specific parsers:

  // perform any necessary initialization (e.g., reading headers)
  // after opening a log file
  virtual bool initialize() { return true; }

  // any necessary cleanup
  virtual void cleanup() {}

  // read a single line from the logfile and update pose/scan as
  // necessary
  virtual datatype_t read_datum() = 0;

};

// automatically open the right kind of log file based on the
// extension
logfile_t * autoopen_log(const char *filename);

// tokenize line by whitespace separators
int tokenize(char *line, std::vector<char *> &tokens);


// reader for data files in the RPI ASCII log format
class log_rpi : public logfile_t
{
public:
  log_rpi() : logfile_t(), offset_x(0), offset_y(0), offset_z(0) {}
  virtual bool initialize();
  virtual datatype_t read_datum();
protected:
  double offset_x, offset_y, offset_z; // offset of laser source from odometry pose
};

// reader for data files in the RPI binary log format
class log_rpi_bin : public log_rpi
{
public:
  log_rpi_bin() : log_rpi() {}
  virtual bool initialize();
  virtual datatype_t read_datum();
};

// reader for data files in the Player/Stage log format
class log_player : public logfile_t
{
public:
  log_player() : logfile_t() {}
  virtual datatype_t read_datum();
};

// reader for data files in the CARMEN log format
class log_carmen : public logfile_t
{
public:
  log_carmen() : logfile_t() {}
  virtual datatype_t read_datum();
};

#endif // _LOGFILE_HPP
