/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <stdio.h>
#include "logfile.hpp"

int main(int argc, char **argv)
{
  if(argc < 2) {
    printf("Usage: %s logfile\n", argv[0]);
    return 1;
  }

  logfile_t *logfile = autoopen_log(argv[1]);
  if(!logfile) {
    printf("Can't open %s\n", argv[1]);
    return 1;
  }

  // first, dump pose history and scan data in separate files

  FILE *ranges = fopen("output.ranges", "w");
  pose_history_t x_1_t;

  fprintf(ranges, "%lg %lg %lg\n", param::scan_min, param::scan_max, param::scan_res);
  while(logfile->read_scan()) {
    x_1_t.push_back(logfile->pose());
    for(uint32_t i = 0; i < logfile->scan().size(); ++i)
      fprintf(ranges, "%lg ", logfile->scan()[i]);
    fprintf(ranges, "\n");
  }

  fclose(ranges);
  x_1_t.write_to_file("output");


  // now, compute the ground truth of (x,y) from gps (assuming we have
  // it -- if not, this file will be full of zeros)
  logfile->reopen();
  logfile_t::datatype_t type;
  std::vector<double> odom;
  typedef std::pair<double, point_t> gps_pair;
  std::vector<gps_pair> gps;
  while(!logfile->eof()) {
    type = logfile->read_datum();
    if(type == logfile_t::ERROR)
      break;
    else if(type == logfile_t::GPS)
      gps.push_back(gps_pair(logfile->time(), logfile->gps()));
    else if(type == logfile_t::ODOMETRY)
      odom.push_back(logfile->time());
  }

  // compute gps poses at the times of all the odometry estimates by
  // interpolating (linearly) bewteen the last and next gps poses.  if
  // the time difference between the last and next gps poses is too
  // great, set the ground truth pose at that time to (NaN,NaN).  for
  // gps, we will always set theta = NaN since we have no heading
  // information.
  double last = -max_double();
  uint32_t gps_cur = 0;
  double next = gps[gps_cur].first;
  pose_history_t gps_truth;
  for(uint32_t i = 0; i < odom.size(); ++i) {
    // advance until odom[i] is between last and next times
    while(next < odom[i]) {
      last = next;
      if(gps_cur+1 == gps.size())
	next = max_double();
      else
	next = gps[++gps_cur].first;
    }
    if(next - last > 0.5)
      gps_truth.push_back(pose_t(NAN,NAN,NAN));
    else {
      double w = (odom[i]-last)/(next-last);
      point_t p = w*gps[gps_cur].second + (1.0-w)*gps[gps_cur-1].second;
      gps_truth.push_back(pose_t(p.x(),p.y(),NAN));
    }
  }
  gps_truth.write_to_file("gps-truth");

  return 0;
}

// hack: put this here so we can link with logfile.cpp
namespace param
{
  // filters
  uint32_t N; // # particles
  uint32_t m; // scans per multiscan
  slam_algorithm slam_alg;
  resample_algorithm resample_alg;
  double resample_neff_percent;
  double resample_alpha; // only used for generalized resampling
  uint32_t lag_time; // for fixed-lag filters

  // model
  pose_t origin;
  double scan_min, scan_max, scan_res;
  uint32_t sparse_points;
  double max_range;
  double err_range, err_bearing, err_trans, err_rot, err_norm;

  // feature extraction
  double le_min_length;
  uint32_t le_min_points;
  double le_merge_threshold;
  double le_converge_threshold;
  double le_tmax;
  double le_cluster_threshold;

  // data association
  double da_ml_threshold;

  // misc
  bool gui;
  bool known_features;
  FILE *sample_dist_log;
  FILE *full_sample_dist_log;
  FILE *neff_log;
  FILE *performance_log;
};
