/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// split-merge that requires an ordering of points --- useful for
// e.g., laser data

#include "slam.hpp"

double tmax_2, clust_thresh_2, min_len_2, sd2, sp2;


enum fit_type {
  LEAST_SQUARES,
  WEIGHTED_ML
};

struct range
{
  uint32_t start, end;
  range() {}
  range(uint32_t s, uint32_t e) : start(s), end(e) {}
  uint32_t size() const { return end - start; }
};

// simple unweighted least squares estimation (no covariance
// estimation); borrowed mainly from GSL, aside from the
// transformation at the end.  used for the iterative segmentation
// step since it's much faster than full ML estimation
void least_squares_fit
  (const cartesian_scan_t &points, const range &r, line_feature &out)
{
  double m_x = 0, m_y = 0, m_dx2 = 0, m_dxdy = 0;
  uint32_t i, j;

  for(i = r.start, j = 0; i < r.end; ++i, ++j) {
    m_x += (points[i].x() - m_x) / (j + 1.0);
    m_y += (points[i].y() - m_y) / (j + 1.0);
  }

  for(i = r.start, j = 0; i < r.end; ++i, ++j) {
    const double dx = points[i].x() - m_x;
    const double dy = points[i].y() - m_y;
    m_dx2 += (dx * dx - m_dx2) / (j + 1.0);
    m_dxdy += (dx * dy - m_dxdy) / (j + 1.0);
  }

  // y=mx+b form
  double m = m_dxdy / m_dx2;
  double b = m_y - m_x * m;

  // transform to r=x*cos(theta) + y*sin(theta) form
  out.r() = -sign(m)*b/sqrt(1+m*m); // r
  out.t() = -atan(1/m); // theta
}

// Line estimation from polar range-bearing data, based mainly on:
// "Weighted line fitting algorithms for mobile robot map building and
// efficient data representation."  Pfister, Roumeliotis, Burdick,
// ICRA03.  In general, when using data from multiple poses we
// (technically) can't apply this estimator because it assumes
// ordered, radial data from a single pose.  However, if the data is
// from consecutive poses over a reasonably short interval the
// approximation isn't bad.  (The approximation is to assume all the
// data was actually taken from a single pose -- e.g., the most recent
// pose -- and compute range-bearing data from the raw cartesian
// points under this assumption.)  The advantage is simplicity and
// efficiency.
void ml_polar_line_estimate
  (const cartesian_scan_t &points, const range &r, line_feature &out)
{
  uint32_t k, j;

  // first get polar versions of all the points.  this is far from
  // ideal but it turns out to be an easy thing to do that works
  // fairly well
  static std::vector<double> phi(1024), d(1024); // pre-allocation for speed
  phi.resize(r.size()); d.resize(r.size());
  for(k = r.start, j = 0; k < r.end; ++k, ++j) {
    const point_t &p = points[k];
    phi[j] = atan2(p.y(), p.x());
    d[j] = ::sqrt(p.x()*p.x()+p.y()*p.y());
  }

  // get an initial estimate of the line parameters
  least_squares_fit(points, r, out);

  // static for speed
  static double ck, sk, da, Pk, Prr, Rsum, ak, akp, akpp, bk, bkp, bkpp, num, GTpp;

  bool done = false;

  do {
    // compute r
    Prr = 0;
    Rsum = 0;
    for(k = 0; k < d.size(); ++k) {
      const double &dk = d[k];
      ck = cos(out.t() - phi[k]);
      sk = sin(out.t() - phi[k]);
      Pk = sd2*ck*ck + sp2*dk*dk*sk*sk;
      Prr += 1 / Pk;
      Rsum += dk*ck / Pk;
    }
    Prr = 1 / Prr;
    out.r() = Prr * Rsum;

    // update theta estimate
    num = 0;
    GTpp = 0;
    for(k = 0; k < d.size(); ++k) {
      const double &dk = d[k];
      ck = cos(out.t() - phi[k]); // FIXME store these somehow to avoid recomputing?
      sk = sin(out.t() - phi[k]);
      ak = dk*ck - out.r(); ak*=ak;
      akp = -2*dk*sk*(dk*ck - out.r());
      akpp = 2*dk*dk*sk*sk - 2*dk*ck*(dk*ck - out.r());
      bk = sd2*ck*ck + sp2*dk*dk*sk*sk;
      bkp = 2*ck*sk*(dk*dk*sp2 - sd2);
      bkpp = 2*(dk*dk*sp2 - sd2)*(ck*ck - sk*sk);
      num += (bk*akp-ak*bkp)/(bk*bk);
      GTpp += ((akpp*bk-ak*bkpp)*bk - 2*(akp*bk-ak*bkp)*bkp) / (bk*bk*bk);
    }

    if(done)
      break;

    da = num/GTpp;
    out.t() -= da;

    // if we are done, we'll repeat the computation of GTpp one more
    // time so that we can use the final value below
    if(fabs(da) <= param::le_converge_threshold)
      done = true;
  } while(1);

  // compute covariance
  out.cov = vl_zero;

  // Prr
  Prr = 0;
  for(k = 0; k < d.size(); ++k) {
    const double &dk = d[k];
    ck = cos(out.t() - phi[k]);
    sk = sin(out.t() - phi[k]);

    bk = sd2*ck*ck + sp2*dk*dk*sk*sk;
    Prr += 1 / bk;
    out.cov[0][1] += 2*dk*sk / bk;
    out.cov[1][1] += 4*dk*dk*sk*sk / bk;
  }
  out.cov[0][0] = 1 / Prr;
  out.cov[0][1] *= (out.cov[0][0] / GTpp);
  out.cov[1][0] = out.cov[0][1];
  out.cov[1][1] /= GTpp * GTpp;

  out.cov *= 10; // FIXME remove?
  //  out.cov.rr *= 50; out.cov.rt *= 50; out.cov.tt *= 50;
}

// FIXME remove
void test_ml_polar_esitmate()
{
  sd2 = param::err_range * param::err_range; // var of range measurement
  sp2 = param::err_bearing * param::err_bearing; // var of bearing measurement
  cartesian_scan_t foo;
  foo.push_back(point_t(-0.5,0.5));
  foo.push_back(point_t(0,0.5));
  foo.push_back(point_t(0.5,0.5));
  range r(0, 3);
  line_feature l;
  ml_polar_line_estimate(foo, r, l);
  //  least_squares_fit(foo, r, l);
  printf("line: %f %f [%f %f %f]\n", l.r(), l.t(), l.cov[0][0], l.cov[0][1], l.cov[1][1]);
  // according to matlab this should yield: -0.500000 -1.570796 [0.000500 0.000000 0.002500]
  exit(0);
}


void fit
  (std::vector<extent_feature> &out,
   const cartesian_scan_t &points,
   const std::vector<range> &ranges,
   fit_type method)
{
  out.resize(ranges.size());
  double u;
  for(uint32_t s = 0; s < out.size(); ++s) {
    // estimate line
    if(method == LEAST_SQUARES)
      least_squares_fit(points, ranges[s], out[s]);
    else
      ml_polar_line_estimate(points, ranges[s], out[s]);

    // truncate
    out[s].umax = -max_double();
    out[s].umin = max_double();
    for(uint32_t i = ranges[s].start; i < ranges[s].end; ++i) {
      u = project_onto(points[i], out[s]);
      if(u < out[s].umin) out[s].umin = u;
      if(u > out[s].umax) out[s].umax = u;
    }
  }
}

void split
  (const std::vector<extent_feature> &out,
   const cartesian_scan_t &points,
   std::vector<range> &working_ranges,
   std::vector<range> &done_ranges)
{
  std::vector<range> newranges;
  double d;
  for(uint32_t s = 0; s < out.size(); ++s) {
    const range &wr = working_ranges[s];

    // first find the farthest point from the segment
    double maxd = -max_double();
    uint32_t maxi = wr.start;
    for(uint32_t i = wr.start; i < wr.end; ++i) {
      d = squared_distance(points[i], out[s]);
      if(d > maxd) {
	maxd = d;
	maxi = i;
      }
    }

    if(maxd < tmax_2) // done with this segment
      done_ranges.push_back(wr);
    else { // split into two segments
      if(maxi == wr.start)
	++maxi;
      else if(maxi == wr.end)
	--maxi;

      if(maxi - wr.start >= param::le_min_points)
	newranges.push_back(range(wr.start, maxi));
      if(wr.end - maxi >= param::le_min_points)
	newranges.push_back(range(maxi, wr.end));
    }
  }

  working_ranges = newranges;
}

void merge(std::vector<extent_feature> &segments)
{
  uint32_t s, t;
  double lh;

  // initialization
  double ml[segments.size()];
  uint32_t c[segments.size()];
  for(s = 0; s < segments.size(); ++s) {
    ml[s] = 0;
    c[s] = s;
  }

  // compute pairwise matching likelihoods of segments that overlap
  for(s = 0; s < segments.size(); ++s) {
    for(t = s+1; t < segments.size(); ++t) {
      if(overlap(segments[s], segments[t])) {
	lh = equality_likelihood(segments[s], segments[t]);
	if(lh > ml[s]) { ml[s] = lh; c[s] = t; }
	if(lh > ml[t]) { ml[t] = lh; c[t] = s; }
      }
    }
  }

  // maximum likelihood pairings with likelihood > thresh go in the
  // same cluster; all segments in a single cluster get merged into a
  // single segment
  std::vector<extent_feature> out;
  for(s = 0, t = 0; s < segments.size(); ++s) {
    if(c[s] < s && ml[s] > param::le_merge_threshold) {
      c[s] = c[c[s]];
      out[c[s]] = merge(out[c[s]], segments[s]);
    } else {
      c[s] = t++;
      out.push_back(segments[s]);
    }
  }

  segments = out;
}


// assumes points are in "order" (typically radial wrt pose)
void extract_features
  (const cartesian_scan_t &points, std::vector<extent_feature> &out)
{
  out.clear();
  std::vector<range> working_ranges, done_ranges;
  std::vector<extent_feature> segments;

  // initialize some parameters
  // FIXME: do this in some sort of call-once initialization?
  tmax_2 = param::le_tmax * param::le_tmax;
  clust_thresh_2 = param::le_cluster_threshold * param::le_cluster_threshold;
  min_len_2 = param::le_min_length * param::le_min_length;
  sd2 = param::err_range * param::err_range; // var of range measurement
  sp2 = param::err_bearing * param::err_bearing; // var of bearing measurement

  // do initial clustering (breakpoint detection)
  uint32_t last = 0, i;
  for(i = 1; i < points.size(); ++i) {
    if(squared_distance(points[i], points[i-1]) > clust_thresh_2) {
      if(i - last > param::le_min_points)
	working_ranges.push_back(range(last, i));
      last = i;
    }
  }
  if(i - last > param::le_min_points)
    working_ranges.push_back(range(last, i));

  // iteratively fit, split, repeat until converged
  while(working_ranges.size()) {
    fit(segments, points, working_ranges, LEAST_SQUARES);
    split(segments, points, working_ranges, done_ranges);
  }

  // do full ML estimation of parameters and covariances
  fit(segments, points, done_ranges, WEIGHTED_ML);

  // merge similar segments
  merge(segments);

  // discard too-short segments, return the rest
  for(i = 0; i < segments.size(); ++i)
    if(squared_length(segments[i]) >= min_len_2)
      out.push_back(segments[i]);
}
