/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// RBPF SLAM with fixed-lag roughening

#include "rbpf.hpp"


slam_rbpf_roughening::slam_rbpf_roughening(const pose_t &start_pose)
  : slam_rbpf_improved<lag_particle_t>(start_pose)
{
  map_type m;
  std::vector<uint32_t> n(0);
  fixed_lag_data blank;
  blank.u = pose_t(0,0,0);
  blank.u_ref_frame = blank.z_ref_frame = pose_t(0,0,0);
  for(uint32_t i = 0; i < param::N; ++i) {
    lag_particle_t &p = particles[i];
    fixed_lag_push_front(p.poses, start_pose);
    fixed_lag_push_front(p.maps, m);
    fixed_lag_push_front(p.n, n);
  }
  fixed_lag_push_front(lag_data, blank);
}

void slam_rbpf_roughening::add_features
  (const std::vector<feature_type> &features,
   const std::vector<uint32_t> &associations,
   const pose_t &ref_frame)
{
  // make space for new data associations computed in rbpf_improved's
  // add_features
  std::vector<uint32_t> tmp_n(features.size());
  for(uint32_t i = 0; i < particles.size(); ++i)
    fixed_lag_push_front(particles[i].n, tmp_n); // empty for now

  // first do a normal improved proposal update
  slam_rbpf_improved<lag_particle_t>::internal_add_features
    (features, associations, ref_frame);

  for(uint32_t i = 0; i < particles.size(); ++i) {
    lag_particle_t &p = particles[i];
    fixed_lag_push_front(p.poses, p.pose);
    fixed_lag_push_front(p.maps, p.get_map()); // copy the last map
  }

  // save the data from this frame for later use in roughening
  fixed_lag_data d;
  d.u = last_delta;
  d.u_ref_frame = last_ref_frame;
  d.z = features;
  d.z_ref_frame = ref_frame;
  fixed_lag_push_front(lag_data, d);

  //////////////////////////////////////////////////////////////////////
  // now for the move step
  //////////////////////////////////////////////////////////////////////

  for(uint32_t i = 0; i < particles.size(); ++i) {
    lag_particle_t &p = particles[i];
    for(int k = param::lag_time-1; k >= 0; --k) // gibbs sampling
      fixed_lag_sample(p, k);
    // change the pose histories
    p.pose = p.poses.front();
    p.history.fixed_lag_modify(p.poses);
  }

  do_analysis();
}

// k is relative to t (i.e., sample x_{t-k}); this should be called
// with DECREASING values of k
void slam_rbpf_roughening::fixed_lag_sample(lag_particle_t &p, uint32_t k)
{
  assert(k < param::lag_time);

  if(int(lag_data.size())-2 < int(k)) // -2: never change the first pose, and k is 0-indexed
    return;

  // initialize a bunch of iterators for going through the fixed-lag data
  std::list<map_type>::iterator mi = p.maps.begin(), m_k;
  std::list<pose_t>::iterator pi = p.poses.begin(), p_k;
  std::list< std::vector<uint32_t> >::iterator ni = p.n.begin(), n_k;
  std::list<fixed_lag_data>::iterator di = lag_data.begin(), d_k;

  // FIXME check associations here

  for(uint32_t l = 0; l < k; ++l, ++mi, ++pi, ++di, ++ni);
  m_k = mi; p_k = pi; n_k = ni; d_k = di;
  assert(mi != p.maps.end());
  map_type map_k = *(++mi); // this map incorporates all measurements up to time t-k-1

  // compute the forward model from u_{t-k} and x_{t-k-1}
  pose_t mu; cov3_t cov(cov_epsilon); // see rbpf.hpp
  assert(pi != p.poses.end()); // should be one extra pose stored!
  motion_model(di->u, di->u_ref_frame, *(++pi), mu, cov); //++pi: t-k-1

  double dummy_weight = 0;
  if(k > 0) { // don't do this if we're at time t
    // compute the backward model from u_{t-k+1} and x_{t-k+1} and
    // merge it with the forward model
    --di; --pi; --pi; --mi; --ni; // t-k+1
    pose_t mu_b(-di->u.x(), -di->u.y(), -di->u.t()); cov3_t cov_b = cov_epsilon;
    motion_model(mu_b, di->u_ref_frame, *pi, mu_b, cov_b);

    merge(mu, cov, mu_b, cov_b);

    // now update the map conditioned on poses t-k+1..t and the stored
    // measurements and associations
    do {
      std::vector<feature_type> xf_features(di->z.size());
      for(uint32_t f = 0; f < xf_features.size(); ++f) // transform features to *pi frame
	feature_to_particle_frame(di->z[f], di->z_ref_frame, xf_features[f], *pi);
      update_map_improved(map_k, xf_features, *ni, *pi, dummy_weight);
      if(di == lag_data.begin())
	break;
      --di; --pi; --mi; --ni;
    } while(1);
  }

  // FIXME: should we recompute the data associations using the
  // full-information map?  (in practice it probably doesn't make a
  // big difference but from a theoretical standpoint, "yes" is most
  // likely the correct answer.)

  // finally compute the improved proposal distribution using the
  // observations z_{t-k} and associations n_{t-k} with the map we
  // just computed
  map_type::iterator assoc;
  feature_type feat;
  for(uint32_t f = 0; f < d_k->z.size(); ++f) {
    assoc = map_k.find_id((*n_k)[f]);
    if(assoc) {
      feature_to_particle_frame(d_k->z[f], d_k->z_ref_frame, feat, mu);
      improve_proposal(mu, cov, feat, assoc->landmark);
    }
  }

  // sample x_{t-k} from the computed distribution
  *p_k = gaussian(mu, cov);

  // apply the usual map update to M_{t-k-1} to obtain M_{t-k}
  // conditioned on the new sampled value
  mi = m_k; ++mi; *m_k = *mi; // set M_{t-k}=M_{t-k-1}

  std::vector<feature_type> xf_features(d_k->z.size()); // transform features to *p_k frame
  for(uint32_t f = 0; f < d_k->z.size(); ++f)
    feature_to_particle_frame(d_k->z[f], d_k->z_ref_frame, xf_features[f], *p_k);
  update_map_improved(*m_k, xf_features, *n_k, *p_k, dummy_weight, &mu, &cov);

  // note that we don't update the weight of the particle
}
