/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _RESAMPLE_HPP
#define _RESAMPLE_HPP

// variations on resampling including simple random resampling,
// residual resampling, and generalized resampling

// all of these algorithms assume the weights have already been
// normalized!

// slam algorithms should just call resample, which will pick the
// right algorithm based on command line options

#include "slam.hpp"


// simple random resampling: draw N particles with replacement
// according to the weights of the particles

template <class Particle>
void resample_random(std::vector<Particle> &particles)
{
  std::vector<Particle> new_particles;
  new_particles.reserve(particles.size());

  double r, s;
  uint32_t i, j;
  for(i = 0; i < particles.size(); ++i) {
    r = drand48();
    for(j = 0, s = 0; j < particles.size(); ++j) {
      s += particles[j].w;
      if(s > r) break;
    }
    new_particles.push_back(particles[j]);
    new_particles.back().w = 1.0/particles.size();
  }

  particles = new_particles;
}


// residual resampling: pick a number of copies of each particle
// proportional to its weight (Liu, 2001)

template <class Particle>
void resample_residual(std::vector<Particle> &particles)
{
  std::vector<Particle> new_particles;
  new_particles.reserve(particles.size());

  double W = 0;
  uint32_t i, j, Nr;
  for(i = 0; i < particles.size(); ++i) {
    j = (uint32_t)floor(particles.size()*particles[i].w);
    for(; j > 0; --j) {
      new_particles.push_back(particles[i]);
      new_particles.back().w = 1.0/particles.size();
    }
    particles[i].w = particles[i].w*particles.size() - j; // residual
    W += particles[i].w;
  }
  for(i = 0; i < particles.size(); ++i)
    particles[i].w /= W; // normalize residual weights

  Nr = particles.size()-new_particles.size();

  double r, s;
  for(i = 0; i < Nr; ++i) {
    r = drand48();
    for(j = 0, s = 0; j < particles.size(); ++j) {
      s += particles[j].w;
      if(s > r) break;
    }
    new_particles.push_back(particles[j]);
    new_particles.back().w = 1.0/particles.size();
  }

  particles = new_particles;
}


// generalized random resampling: random resampling with weights that
// are a function of the importance weights of the particles; in this
// particular implementation, draw particles according to a_i =
// w_i^{\alpha}, 0 < \alpha <= 1 (typically, \alpha = 1/2 is a good
// choice) (Liu 2001); then, reset weights to w_i/a_i instead of 1/N.

template <class Particle>
void resample_generalized(std::vector<Particle> &particles)
{
  std::vector<double> a(particles.size());
  std::vector<Particle> new_particles;
  new_particles.reserve(particles.size());

  uint32_t i, j;
  double r, s, a_sum = 0;
  for(i = 0; i < particles.size(); ++i) {
    a[i] = pow(particles[i].w, param::resample_alpha);
    a_sum += a[i];
  }
  for(i = 0; i < particles.size(); ++i)
    a[i] /= a_sum; // normalize generalized weights

  for(i = 0; i < particles.size(); ++i) {
    r = drand48();
    for(j = 0, s = 0; j < particles.size(); ++j) {
      s += a[j];
      if(s > r) break;
    }
    new_particles.push_back(particles[j]);
    new_particles.back().w /= a[j];
  }

  particles = new_particles;
}


// generic resample, pick the right algorithm

template <class Particle>
void resample(std::vector<Particle> &particles)
{
  switch(param::resample_alg) {
  case RANDOM: resample_random(particles); break;
  case RESIDUAL: resample_residual(particles); break;
  case GENERALIZED: resample_generalized(particles); break;
  }
}

#endif // _RESAMPLE_HPP
