/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _SAMPLE_DISTRIBUTION_HPP
#define _SAMPLE_DISTRIBUTION_HPP

#include "slam.hpp"

// compute the (weighted) sample mean and covariance for the robot's
// current pose from the current particle set; if
// param::sample_dist_log is a valid file, write the estimated
// distribution to it for use in examining NEES
template <class Particle>
void sample_distribution
  (const std::vector<Particle> &particles, FILE *log)
{
  if(!log)
    return;

  pose_t mean(vl_0);
  double xx=0, xy=0, xt=0, yy=0, yt=0, tt=0;
  double W = 0; // total weight
  double sinm = 0, cosm = 0; // average angles using vectors
  uint32_t i;

  // sample mean
  for(i = 0; i < particles.size(); ++i) {
    const Particle &p = particles[i];
    mean.x() += p.pose.x() * p.w;
    mean.y() += p.pose.y() * p.w;
    sinm += sin(p.pose.t()) * p.w;
    cosm += cos(p.pose.t()) * p.w;
    W += p.w;
  }
  mean.t() = clamp_angle(atan2(sinm/W, cosm/W));
  mean.x() /= W;
  mean.y() /= W;

  // sample covariance
  for(i = 0; i < particles.size(); ++i) {
    const Particle &p = particles[i];
    const double dx = p.pose.x() - mean.x();
    const double dy = p.pose.y() - mean.y();
    const double dt = angular_distance(p.pose.t(), mean.t());
    xx += dx * dx * p.w / W;
    xy += dx * dy * p.w / W;
    xt += dx * dt * p.w / W;
    yy += dy * dy * p.w / W;
    yt += dy * dt * p.w / W;
    tt += dt * dt * p.w / W;
  }

  // FIXME: should not output degrees because covariance is for radians!
  fprintf(log, "%g %g %g %g %g %g %g %g %g %g %g %g\n",
	  mean.x(), mean.y(), rad2deg(mean.t()),
	  xx, xy, xt, xy, yy, yt, xt, yt, tt);
}


// compute the weighted sample mean and covariance over the entire
// pose history of the robot (instead of just over the current pose)
// which can reveal a lot about particle degeneracy; only call this
// infrequently (e.g., at the end of slam) because it's expensive
struct simple_particle_t { double w; pose_t pose; };
template <class Particle>
void full_sample_distribution
  (const std::vector<Particle> &particles, FILE *log)
{
  if(!log)
    return;

  std::vector< std::list<pose_t> > ph(particles.size());
  std::vector<simple_particle_t> simp_parts(particles.size());
  uint32_t i;
  for(i = 0; i < particles.size(); ++i) {
    particles[i].history.get_stl_list(ph[i]);
    simp_parts[i].w = particles[i].w; // use final weights
  }

  while(ph[0].size() != 0) {
    for(i = 0; i < particles.size(); ++i) {
      simp_parts[i].pose = ph[i].front(); // set particles to time j pose
      ph[i].pop_front();
    }
    sample_distribution(simp_parts, log);
  }
}

#endif // _SAMPLE_DISTRIBUTION_HPP
