/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// scan-matching slam with occupancy maps

#include "slam.hpp"
#include "scan_match.hpp"
#include "neff.hpp"
#include "sample_distribution.hpp"
#include "resample_random.hpp"
#include <GL/glut.h>

// FIXME: make these parameters
grid_map_t::grid_map_t()
  : xmin(-40), ymin(-40), xmax(40), ymax(40), res(10),
    width(xmax-xmin), height(ymax-ymin),
    rows((uint32_t)(width * res)), cols((uint32_t)(height * res)),
    size(rows * cols)
{
  grid = new int8_t[size];
  memset(grid, 0, size);
}

grid_map_t & grid_map_t::operator=(const grid_map_t &m)
{
  if(grid)
    delete [] grid;
  xmin = m.xmin; ymin = m.ymin; xmax = m.xmax; ymax = m.ymax;
  res = m.res; width = m.width; height = m.height;
  rows = m.rows; cols = m.cols; size = m.size;
  grid = new int8_t[size]; memcpy(grid, m.grid, size);
  return *this;
}

inline void grid_map_t::rowcol(double x, double y, uint32_t &r, uint32_t &c)
{
  r = (uint32_t)floor((x-xmin)*res);
  c = (uint32_t)floor((y-ymin)*res);
}

inline uint32_t grid_map_t::cell(uint32_t r, uint32_t c)
{
  if(r >= rows || c >= cols)
    return max_uint32();
  return (rows-r-1) * cols + c;
}

inline void grid_map_t::occupied(uint32_t cell)
{
  if(cell < size)
    grid[cell] = std::min(127, grid[cell] + 1);
}

inline void grid_map_t::unoccupied(uint32_t cell)
{
  if(cell < size)
    grid[cell] = std::max(-127, grid[cell] - 1);
}

inline double grid_map_t::get_likelihood
  (const point_t &p, uint32_t r, uint32_t c,
   point_t &gp, double &d)
{
  gp.x() = r/res+xmin;
  gp.y() = c/res+ymin;
  d = squared_distance(p, gp);
  return exp(-d) * (grid[cell(r,c)]+127) / 255.0;
}

// find the best matching grid cell in the map for the point p,
// according to the likelihood: exp(-sqdist(cell_center, point)) *
// cell_likelihood
void grid_map_t::match_single_point
  (const point_t &p, double &likelihood, uint32_t &r, uint32_t &c)
{
  rowcol(p.x, p.y, r, c);
  double l, d;
  uint32_t rb = r-1, cb = c-1;
  point_t gp;

  likelihood = get_likelihood(p, rb, cb, gp, d);

  l = get_likelihood(p, r, c-1, gp, d);
  if(l > likelihood) { rb = r; likelihood = l; }
  l = get_likelihood(p, r+1, c-1, gp, d);
  if(l > likelihood) { rb = r; likelihood = l; }
  l = get_likelihood(p, r-1, c, gp, d);
  if(l > likelihood) { rb = r-1; cb = c; likelihood = l; }
  l = get_likelihood(p, r, c, gp, d);
  if(l > likelihood) { rb = r; cb = c; likelihood = l; }
  l = get_likelihood(p, r+1, c, gp, d);
  if(l > likelihood) { rb = r+1; cb = c; likelihood = l; }
  l = get_likelihood(p, r-1, c+1, gp, d);
  if(l > likelihood) { rb = r-1; cb = c+1; likelihood = l; }
  l = get_likelihood(p, r, c+1, gp, d);
  if(l > likelihood) { rb = r; cb = c+1; likelihood = l; }
  l = get_likelihood(p, r+1, c+1, gp, d);
  if(l > likelihood) { rb = r+1; cb = c+1; likelihood = l; }
  r = rb;
  c = cb;
}

// the main scan matching function, also updates the grid
double grid_map_t::get_likelihood_and_update
  (const cartesian_scan_t &points, const pose_t &pose)
{
  // for each point, find a cell in its neighborhood in the map with
  // the highest likelihood, computed as exp(-sqdist(cell_center,
  // point)) * cell_likelihood; if this likelihood is large enough,
  // increase occupancy for that cell, otherwise update the map based
  // on the point location

  // note that this isn't as good as doing some kind of image
  // registration based technique since it ignores the jointness of
  // the data association

  double likelihood = 1, lh;
  uint32_t r, c;
  for(uint32_t i = 0; i < points.size(); ++i) {
    match_single_point(points[i], lh, r, c);
    if(lh > param::da_ml_threshold) { // FIXME this variable will change?
      likelihood *= lh;
      occupied(cell(r,c));
    } else {
      likelihood *= param::da_ml_threshold;
      update(points[i], pose);
    }
  }
  return likelihood;
}

// same as above, without any updating
double grid_map_t::get_likelihood(const cartesian_scan_t &points)
{
  double likelihood = 1, lh;
  uint32_t r, c;
  for(uint32_t i = 0; i < points.size(); ++i) {
    match_single_point(points[i], lh, r, c);
    if(lh > param::da_ml_threshold) // FIXME this variable will change?
      likelihood *= lh;
    else
      likelihood *= param::da_ml_threshold;
  }
  return likelihood;
}


// fast bresenham's algorithm, adapted from
// http://www.cs.unc.edu/~mcmillan/comp136/Lecture6/Lines.html
void grid_map_t::bresenham(int x0, int y0, int x1, int y1)
{
  int dy = y1 - y0;
  int dx = x1 - x0;
  int stepx, stepy;

  if(dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
  if(dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
  dy <<= 1;
  dx <<= 1;

  unoccupied(cell(x0, y0));
  if (dx > dy) {
    int fraction = dy - (dx >> 1);
    while(x0 != x1) {
      if(fraction >= 0) {
	y0 += stepy;
	fraction -= dx;
      }
      x0 += stepx;
      fraction += dy;
      unoccupied(cell(x0, y0));
    }
  } else {
    int fraction = dx - (dy >> 1);
    while(y0 != y1) {
      if(fraction >= 0) {
	x0 += stepx;
	fraction -= dy;
      }
      y0 += stepy;
      fraction += dx;
      unoccupied(cell(x0, y0));
    }
  }

  // twice since we already did unoccupied in this cell once
  occupied(cell(x1, y1));
  occupied(cell(x1, y1));
}

//#if 0
void grid_map_t::update(const point_t &p, const pose_t &pose)
{
  uint32_t x0, y0, x1, y1;
  rowcol(pose.x(), pose.y(), x0, y0);
  rowcol(p.x(), p.y(), x1, y1);
  bresenham(x0, y0, x1, y1);
}
//#endif

#if 0
void grid_map_t::update(const point_t &p, const pose_t &pose)
{
  uint32_t count, cxmin, cymin, cxmax, cymax, rng, r, c;

  // increase occupancy for the cell containing the point
  rowcol(p.x(), p.y(), r, c);
  occupied(cell(r,c));

  // decrease occupancy for cells along the line between the robot and
  // the point
  count = (uint32_t)floor(distance(p, pose)) * 5; // FIXME?
  double dx = (p.x() - pose.x()) / count;
  double dy = (p.y() - pose.y()) / count;
  cxmin = cymin = size;
  rowcol(p.x(), p.y(), cxmax, cymax);
  point_t curpt(pose.x(), pose.y());
  for(; count > 0; --count) { // step along line
    rowcol(curpt.x(), curpt.y(), r, c);
    rng = cell(r, c);
    // if we haven't updated this cell, and it's not the one with the
    // feature in it, update it
    if(rng < size && (r != cxmin || c != cymin) && (r != cxmax && c != cymax))
      unoccupied(rng);
    curpt.x() += dx;
    curpt.y() += dy;
    cxmin = r;
    cymin = c;
  }
}
#endif

void grid_map_t::render() const
{
  // invert colors (we want white == empty, black == occupied)
  glPixelTransferf(GL_RED_SCALE, -1);
  glPixelTransferf(GL_GREEN_SCALE, -1);
  glPixelTransferf(GL_BLUE_SCALE, -1);

  // cell value 0 is gray
  glPixelTransferf(GL_RED_BIAS, 0.5);
  glPixelTransferf(GL_GREEN_BIAS, 0.5);
  glPixelTransferf(GL_BLUE_BIAS, 0.5);

  glPixelStorei(GL_UNPACK_ROW_LENGTH, cols);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  // trick: if the raster position is offscreen, we still want to draw
  // the bitmap

  // find the raster position
  GLdouble mview[16], proj[16], x, y, z;
  GLint vport[4];
  glGetDoublev(GL_MODELVIEW_MATRIX, mview);
  glGetDoublev(GL_PROJECTION_MATRIX, proj);
  glGetIntegerv(GL_VIEWPORT, vport);
  gluProject(xmin, ymin, 0, mview, proj, vport, &x, &y, &z);

  // save state
  glPushAttrib(GL_TRANSFORM_BIT | GL_VIEWPORT_BIT);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  // set up drawing parameters
  gluOrtho2D(xmin, xmax, ymin, ymax);
  glRasterPos2d(xmin, ymin);

  // FIXME: this zoom assumes the map width/height are the same as the
  // width/height of the viewport (in world coordinates)
  glPixelZoom(float(vport[2]) / float(cols), float(vport[3])  / float(rows));

  // render pixels
  glDrawPixels(rows, cols, GL_LUMINANCE, GL_BYTE, grid);

  // restore state
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glPopAttrib();
}

void grid_map_t::write_to_file(const char *prefix) const
{
  FILE *file = open_file(prefix, "pgm", "wb");
  if(file) {
    fprintf(file, "P5\n%d\n%d\n255\n", rows, cols);
    uint8_t pix;
    for(uint32_t c = 0; c < size; ++c) {
      pix = 255 - (grid[c] + 127);
      fwrite(&pix, sizeof(pix), 1, file);
    }
  }
}

// copied from rbpf.cpp untouched
slam_scan_match::slam_scan_match(const pose_t &start_pose)
  : particles(param::N), best(0)
{
  for(uint32_t i = 0; i < param::N; ++i) {
    particles[i].pose = start_pose;
    particles[i].history.push_back(start_pose);
    particles[i].w = 1.0/param::N;
  }
}

// copied from rbpf.cpp untouched except to change particle_t to
// grid_particle_t
void slam_scan_match::add_odometry(const pose_t &delta, const pose_t &ref_frame)
{
  pose_t mu;
  cov3_t cov;
  // project all the particles forward
  for(uint32_t i = 0; i < particles.size(); ++i) {
    grid_particle_t &p = particles[i];

    motion_model(delta, ref_frame, p.pose, mu, cov);
    p.pose = gaussian(mu, cov); // sample new pose
    p.history.push_back(p.pose); // update pose history
  }
}

void slam_scan_match::add_scan(const multiscan_t &points)
{
  agglomerate_multiscan(points, all_points);
  all_points.transform(pose_t());
  all_points.frame = pose_t();

  uint32_t i;
  double likelihood, W = 0;

  // iterate over all particles
  for(i = 0; i < particles.size(); ++i) {
    grid_particle_t &p = particles[i];

    // transform to the frame of this particle
    //    all_points.transform(p.pose - all_points.frame);
    //    all_points.frame = p.pose;
    cartesian_scan_t xformed_points(all_points);
    xformed_points.transform(p.pose);

    // do scan-matching, get a likelihood for the match, and update
    // the map
    //    likelihood = p.grid_map.get_likelihood_and_update(all_points, p.pose);
    likelihood = p.grid_map.get_likelihood_and_update(xformed_points, p.pose);
    p.w *= likelihood;
    W += p.w;
  }

  // normalize importance weights and compute "best" particle
  best = 0;
  for(i = 0; i < particles.size(); ++i) {
    particles[i].w /= W;
    if(particles[i].w > particles[best].w)
      best = i;
  }

  double Neff = neff(particles);

  // check if we need to resample, and if so, do it
  if(Neff < param::resample_neff_percent * particles.size()) {
    printf("resampling (Neff = %g)\n", Neff);
    resample(particles);
  }

  sample_distribution(particles);
}


void slam_scan_match::render()
{
  // render the grid map on the bottom
  particles[best].grid_map.render();

  uint32_t i;

  // first, all pose histories
  glLineWidth(1);
  glColor3f(0, 0.5, 0);
  for(i = 0; i < particles.size(); ++i)
    particles[i].history.render();

  const grid_particle_t &bp = particles[best];

  // ... and the best pose history ...
  glLineWidth(1);
  glColor3f(1, 0, 1);
  bp.history.render();

  // ... and the best current pose ...
  glPointSize(8);
  glColor3f(0, 1, 0);
  glBegin(GL_POINTS);
  glVertex2f(bp.pose.x(), bp.pose.y());
  glEnd();
  glBegin(GL_LINES);
  glVertex2f(bp.pose.x(), bp.pose.y());
  glVertex2f(bp.pose.x()+cos(bp.pose.t()), bp.pose.y()+sin(bp.pose.t()));
  glEnd();

  // ... and the scan points ...
  all_points.transform(bp.pose - all_points.frame);
  all_points.frame = bp.pose;
  glColor3f(1, 0, 0);
  glPointSize(2);
  all_points.render();
}

void slam_scan_match::write_to_file(const char *prefix) const
{
  particles[best].grid_map.write_to_file(prefix);
  particles[best].history.write_to_file(prefix);
}


//////////////////////////////////////////////////////////////////////
// version with improved proposal distribution (more or less as
// described in Grisetti, Stachniss, Burgard, "Improving grid-based
// SLAM with Rao-Blackwellized particle filters by adaptive proposals
// and selective resampling," ICRA05, and in Cyrill Stachniss's
// thesis, which seems to have some fixes)
//////////////////////////////////////////////////////////////////////

// in this, we're making an implicit assumption that the data comes as
// odometry/scan pairs --- odometry doesn't accumulate over several
// time steps before we get a scan; so, this is best used with full
// laser data instead of sparse-sensing data.  with some extra
// bookkeeping we could deal with less frequent scans as well.

void slam_scan_match_2::add_odometry(const pose_t &delta, const pose_t &ref_frame)
{
  last_delta = delta;
  last_ref_frame = ref_frame;
}

void slam_scan_match_2::add_scan(const multiscan_t &points)
{
  agglomerate_multiscan(points, all_points);
  all_points.transform(pose_t());
  all_points.frame = pose_t();

  double W = 0, eta;
  uint32_t i, j;

  // improved proposal parameters
  double sx, sy, si, ci;
  double st =
    ::sqrt(param::err_norm*param::err_norm*(last_delta.x()*last_delta.x()+last_delta.y()*last_delta.y()) +
	   param::err_rot*param::err_rot*last_delta.t()*last_delta.t()); // stddev of theta

  const uint32_t proposal_particles = 10; // FIXME make this number a parameter
  pose_t tmp_mu[proposal_particles];
  double likelihood[proposal_particles];
  double dx, dy, dt;

  // iterate over particles
  for(i = 0; i < particles.size(); ++i) {
    grid_particle_t &p = particles[i];

    // FIXME FIXME FIXME: use the proposal function for this since
    // this is kind of wrong!!!

    // first compute the (unimproved) proposal distribution for this
    // particle
    si = sin(p.pose.t()-last_ref_frame.t());
    ci = cos(p.pose.t()-last_ref_frame.t());
    pose_t mu(p.pose.x() + ci*last_delta.x() - si*last_delta.y(),
	      p.pose.y() + si*last_delta.x() + ci*last_delta.y(),
	      p.pose.t() + last_delta.t()); // mean
    sx = fabs(param::err_trans*last_delta.x()); // stddev of x
    sy = fabs(param::err_trans*last_delta.y()); // stddev of y

    // now, compute the improved proposal distribution

    eta = 0; // reset normalization constant
    pose_t improved_mu;
    cov3_t improved_cov;

    // compute the improved proposal mean
    for(j = 0; j < proposal_particles; ++j) {
      pose_t &tmu = tmp_mu[j];
      double &lh = likelihood[j];

      // sample from the unimproved distribution; we can just do it
      // per-element like this since we're assuming a diagonal
      // covariance matrix
      tmu = pose_t(gaussian(mu.x(), sx), gaussian(mu.y(), sy), gaussian(mu.t(), st));

      // FIXME: IT WOULD BE REALLY NICE NOT TO HAVE TO MAKE THIS COPY
      // HERE
      cartesian_scan_t xformed_points(all_points);
      xformed_points.transform(tmu);

      // compute the scan-matching likelihood for this pose
      lh = p.grid_map.get_likelihood(xformed_points);

      // update weighted mean for improved proposal
      improved_mu += lh*tmu;

      eta += lh; // update normalization constant
    }

    // normalize the improved proposal mean
    improved_mu /= eta;

    // compute the improved proposal covariance from the samples
    improved_cov = vl_0;
    for(j = 0; j < proposal_particles; ++j) {
      const pose_t &tmu = tmp_mu[j];
      const double &lh = likelihood[j];

      dx = tmu.x() - improved_mu.x();
      dy = tmu.y() - improved_mu.y();
      //      dt = tmu.t - improved_mu.t; // FIXME?
      dt = angular_distance(tmu.t(), improved_mu.t());

      improved_cov[0][0] += dx*dx*lh/eta;
      improved_cov[0][1] += dx*dy*lh/eta;
      improved_cov[0][2] += dx*dt*lh/eta;
      improved_cov[1][1] += dy*dy*lh/eta;
      improved_cov[1][2] += dy*dt*lh/eta;
      improved_cov[2][2] += dt*dt*lh/eta;
    }
    improved_cov[1][0] = improved_cov[0][1];
    improved_cov[2][0] = improved_cov[0][2];
    improved_cov[2][1] = improved_cov[1][2];

    // okay, we've got the proposal distribution, so now, sample the
    // next pose from it and update the map and particle weight

    p.pose = gaussian(improved_mu, improved_cov);
    p.history.push_back(p.pose);

    // FIXME: IT WOULD BE REALLY NICE NOT TO HAVE TO MAKE THIS COPY
    // HERE
    cartesian_scan_t xformed_points(all_points);
    xformed_points.transform(p.pose);

    // FIXME: use the likelihood returned by this instead of eta when
    // updating w?  (no, say the stachniss papers)
    p.grid_map.get_likelihood_and_update(xformed_points, p.pose);

    p.w *= eta;
    W += p.w;
  }

  // normalize importance weights and compute "best" particle
  best = 0;
  for(i = 0; i < particles.size(); ++i) {
    particles[i].w /= W;
    if(particles[i].w > particles[best].w)
      best = i;
  }

  double Neff = neff(particles);

  // check if we need to resample, and if so, do it
  if(Neff < param::resample_neff_percent * particles.size()) {
    printf("resampling (Neff = %g)\n", Neff);
    resample(particles);
  }

  sample_distribution(particles);
}
