/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// slam driver program

#include "slam.hpp"
#include "rbpf.hpp"
#include "options.hpp"
#include "logfile.hpp"
#include "known_f_da.hpp"
#include <fstream>
#include <time.h>
#include <sys/time.h>
#ifdef GRIDSLAM
#include "scan_match.hpp"
#endif

logfile_t *logfile = 0;
log_known_f_da<feature_type> features_file;
pose_t last_pose;
scan_t sparse_scan;
pose_history_t gps_hist;
bool done = false;

namespace gui { extern void run(int argc, char **argv, const char *filename); };
void process();
int parse_command_line(int argc, char **argv);

#ifdef GRIDSLAM // FIXME deal with these in slam_interface somehow
typedef slam_scan_match mapper_type;
//typedef slam_scan_match_2 mapper_type;
#else
typedef slam_interface mapper_type;
#endif
mapper_type mapper;

// FIXME remove
cartesian_scan_t cscan, tscan;
multiscan_t mscan;


int main(int argc, char **argv)
{
  int inpidx = parse_command_line(argc, argv);
  logfile = autoopen_log(argv[inpidx]);
  if(!logfile) {
    printf("Can't open %s\n", argv[inpidx]);
    return 1;
  }

  // known features (and possibly data associations), e.g., from a
  // simulator?
  if(param::known_features) {
    if(!features_file.open(options::quickget<std::string>("features-file").c_str())) {
      printf("Can't open %s\n", options::quickget<std::string>("features-file").c_str());
      return 1;
    }
  }

  // seed rng
  srand48(time(0));

  // initialize filter
  mapper.initialize(param::origin);
  last_pose = logfile->pose(); // set initial pose

  // decide whether to start a gui
  if(options::quickget<bool>("gui") == true)
    gui::run(argc, argv, argv[inpidx]);
  else while(1)
    process();

  return 0;
}


// FIXME: move multiscan stuff to its own file?
void sparsify(const scan_t &scan)
{
  if(scan.size() == param::sparse_points) {
    sparse_scan = scan;
    return;
  }
  sparse_scan.resize(param::sparse_points);
  sparse_scan[0] = scan[0];
  for(uint32_t i = 1; i < param::sparse_points; ++i)
    sparse_scan[i] = scan[i * scan.size() / (param::sparse_points - 1) - 1];
  param::scan_res = (param::scan_max - param::scan_min) / (param::sparse_points - 1);
}

// get the current clock time in seconds
double get_time_s()
{
  struct timeval tv;
  gettimeofday(&tv, 0);
  return double(tv.tv_sec) + 1e6*double(tv.tv_usec);
}

void process()
{
  uint8_t datum_type;
  pose_t odom;

  static uint32_t scount = 0; // FIXME remove

  static uint32_t count_log = 0;
  static uint32_t count_poses = 0, count_scans = 0;
  static double time_start = get_time_s(), time_slam = 0, time_slam_start;

  while(!done) {
    datum_type = logfile->read_datum();
    ++count_log;
    if(!(count_log % 5))
      printf("L %d  T %f  Neff %f                  \r",
	     logfile->line_number(), logfile->time(), mapper.get_neff());

    if(datum_type == logfile_t::ODOMETRY) {

      // compute odometry difference from last pose and give it to the mapper
      odom = pose_t(logfile->pose().x() - last_pose.x(),
		    logfile->pose().y() - last_pose.y(),
		    angular_distance(logfile->pose().t(), last_pose.t()));
      mapper.add_odometry(odom, last_pose);
      last_pose = logfile->pose();
      ++count_poses;

    } else if(datum_type == logfile_t::GPS) {

      gps_hist.push_back(pose_t(logfile->gps().x(), logfile->gps().y(), 0));

    } else if(datum_type == logfile_t::SCAN) {

      ++count_scans;

      // are features (and data associations, possibly) given to us?
      if(param::known_features) {
	if(!features_file.read_frame()) {
	  printf("ERROR reading features file!\n");
	  return;
	}
	time_slam_start = get_time_s();
	mapper.add_features(features_file.features(), features_file.associations());
	time_slam += get_time_s() - time_slam_start;
	return; // redraw if there's a gui
      } else {
	// need to do feature extraction from scan data

	// full laser data
	mscan.clear();
	mscan.push_back(cartesian_scan_t(logfile->scan()));
	time_slam_start = get_time_s();
	mapper.add_scan(mscan);
	time_slam += get_time_s() - time_slam_start;
	return;
#if 0
	// sparse sensing slam
	sparsify(logfile->scan());
	++scount;
	tscan = cartesian_scan_t(sparse_scan);
	tscan.transform(logfile->pose());
	mscan.push_back(tscan);
	if(scount == param::m) {
	  scount = 0;
	  time_slam_start = get_time_s();
	  mapper.add_scan(mscan);
	  time_slam += get_time_s() - time_slam_start;
	  mscan.clear();
	  return; // redraw if there's a gui
	}
#endif
      }

    }

    if(logfile->eof() || datum_type == logfile_t::ERROR) {
      printf("\n");
      done = true;
      if(param::performance_log) {
	fprintf(param::performance_log, "%d %d %d %g %g %g\n",
		count_log, count_poses, count_scans,
		get_time_s() - time_start, time_slam,
		time_slam / double(count_scans));
      }
      mapper.write_to_file("output");
      close_analysis_logs();
      return;
    }

  }

  if(param::gui)
    usleep(10000);
  else
    exit(0);
}


namespace param
{
  // filters
  uint32_t N; // # particles
  uint32_t m; // scans per multiscan
  slam_algorithm slam_alg;
  resample_algorithm resample_alg;
  double resample_neff_percent;
  double resample_alpha; // only used for generalized resampling
  uint32_t lag_time; // for fixed-lag filters

  // model
  pose_t origin;
  double scan_min, scan_max, scan_res;
  uint32_t sparse_points;
  double max_range;
  double err_range, err_bearing;
  double err_trans, err_rot, err_norm;

  // feature extraction
  double le_min_length;
  uint32_t le_min_points;
  double le_merge_threshold;
  double le_converge_threshold;
  double le_tmax;
  double le_cluster_threshold;

  // data association
  double da_ml_threshold;
  uint32_t spurious_discard_count;

  // constraints
  double rectilinearity_delta;

  // misc
  bool gui;
  bool known_features;
  FILE *sample_dist_log;
  FILE *full_sample_dist_log;
  FILE *neff_log;
  FILE *performance_log;
};

int parse_command_line(int argc, char **argv)
{
  // set up commandline/configuration file options
  options::add<bool>("help", 0, "Print usage information", 0, false, options::nodump);
  options::set_cf_options("config", "c");
  options::add<std::string>("save-config", 0, "Save configuration file", 0, "", options::nodump);

  options::add<bool>("gui", "g", "Display the GUI", "Graphical display", false);
  options::add<double>("workspace-width", "w", "Workspace width (m)", "Graphical display", 80);
  options::add<double>("workspace-height", "h", "Workspace height (m)", "Graphical display", 80);
  options::add<int>("gui-width", 0, "GUI width (pixels)", "Graphical display", 800);
  options::add<int>("gui-height", 0, "GUI height (pixels)", "Graphical display", 800);
  options::add<uint32_t>("redraw-freq", 0, "Scans between re-rendering", "Graphical display", 1);
  options::add<bool>("movie", 0, "Dump movie frames to movie/...", "Graphical display", false);

  options::add<double>("origin-x", 0, "Robot origin x-coordinate (m)", "Model", 0);
  options::add<double>("origin-y", 0, "Robot origin y-coordinate (m)", "Model", 0);
  options::add<double>("origin-theta", 0, "Robot origin theta-coordinate (deg)", "Model", 0);
  options::add<double>("max-range", 0, "Maximum laser range (m)", "Model", 10.0);
  options::add<uint32_t>("sparse-points", 0, "Number of sensors in a sparse scan",
			 "Model", 5);

  options::add<double>("err-range", 0, "Range error (exact, in meters)",
		       "Model", 0.05);
  options::add<double>("err-bearing", 0, "Bearing error (exact, in degrees)",
		       "Model", 0.5);

  options::add<double>("err-translation", 0, "Translation error (percent of odometry measurement)",
		       "Model", 0.12);
  options::add<double>("err-rotation", 0, "Rotation error (percent of odometry measurement)",
		       "Model", 0.24);
  options::add<double>("err-normal", 0, "Normal error (percent of odometry measurement)",
		       "Model", 0.12);

  options::add<uint32_t>("num-particles", "n", "Number of particles", "Filters", 1);
  options::add<uint32_t>("multiscan-size", "m", "Scans per multiscan", "Filters", 30);
  options::add<std::string>("slam-algorithm", "a", "One of fs1,fs2,rough,block (fs2 def.)", "Filters", "fs2");
  options::add<std::string>("resample-algorithm", "r", "One of ran,res,gen (ran def.)", "Filters", "ran");
  options::add<double>("resample-neff-percent", 0, "Percentage of effective particles to trigger resampling",
		       "Filters", 0.5);
  options::add<double>("resample-alpha", 0, "Generalized resampling exponent (0 < alpha <= 1)",
		       "Filters", 0.5);
  options::add<uint32_t>("lag-time", "l", "Lag time (steps)", "Filters", 5);

  options::add<double>("le-min-length", 0, "Minimum line length (m)", "Line extraction", 0.5);
  options::add<uint32_t>("le-min-points", 0, "Minimum number of support points for a line",
			 "Line extraction", 3);
  options::add<double>("le-merge-threshold", 0, "Likelihood threshold for merging nearby segments",
		       "Line extraction", 0.2);
  options::add<double>("le-converge-threshold", 0, "Minimum change in theta estimate for nonlinear line estimator",
		       "Line extraction", 1e-3);
  options::add<double>("le-tmax", 0, "Distance of point from segment that triggers splitting (m)",
		       "Line extraction", 0.09);
  options::add<double>("le-cluster-threshold", 0, "Max. distance between consecutive points in same initial cluster (m)",
		       "Line extraction", 0.55);

  options::add<std::string>("features-file", 0, "Log file containing features and (optionally) data associations",
			    "Input files", "");

  options::add<double>("da-threshold", 0, "Data association threshold value (Mahalanobis distance)",
		       "Data association", 2.0);
  options::add<uint32_t>("spurious-discard-count", 0, "Discard after missing this many more times than observing (0=never)",
			 "Data association", 5);

  options::add<double>("delta", 0, "Delta parameter for rectilinearity constraints (deg)",
		       "Constraints", 10.0);

  // parse commandline (and read config file if it is specified)
  int inpidx = options::parse_cmdline(argc, argv);

  if(inpidx < 0) // some kind of error
    exit(1);    // relevant info printed by getopt

  // print usage information
  if(inpidx >= argc || options::quickget<bool>("help") == true) {
    std::cerr << "Usage: " << argv[0] << " [options] logfile" << std::endl;
    options::print_options(std::cout);
    exit(1);
  }

  // save a config file based on these options?
  std::string cfname = options::quickget<std::string>("save-config");
  if(cfname.length() > 0) {
    std::ofstream conf(cfname.c_str());
    if(!conf)
      std::cerr << "Can't write configuration file " << cfname << std::endl;
    else
      options::dump(conf);
  }

  // save options in variables
  param::N = options::quickget<uint32_t>("num-particles");
  param::m = options::quickget<uint32_t>("multiscan-size");
  param::resample_neff_percent = options::quickget<double>("resample-neff-percent");
  param::resample_alpha = options::quickget<double>("resample-alpha");
  param::lag_time = options::quickget<uint32_t>("lag-time");

  std::string s = options::quickget<std::string>("slam-algorithm");
  if(s == std::string("fs1")) param::slam_alg = FS1;
  else if(s == std::string("fs2")) param::slam_alg = FS2;
  else if(s == std::string("rough")) param::slam_alg = ROUGH;
  else if(s == std::string("block")) param::slam_alg = BLOCK;
  else {
    printf("Unknown SLAM algorithm (must be one of fs1,fs2,rough,block)\n");
    exit(1);
  }

  s = options::quickget<std::string>("resample-algorithm");
  if(s == std::string("gen")) param::resample_alg = GENERALIZED;
  else if(s == std::string("res")) param::resample_alg = RESIDUAL;
  else if(s == std::string("ran")) param::resample_alg = RANDOM;
  else {
    printf("Unknown resample algorithm (must be one of ran,res,gen)\n");
    exit(1);
  }

  param::origin.x() = options::quickget<double>("origin-x");
  param::origin.y() = options::quickget<double>("origin-y");
  param::origin.t() = deg2rad(options::quickget<double>("origin-theta"));
  param::sparse_points = options::quickget<uint32_t>("sparse-points");
  param::max_range = options::quickget<double>("max-range");
  param::err_range = options::quickget<double>("err-range");
  param::err_bearing = deg2rad(options::quickget<double>("err-bearing"));
  param::err_trans = options::quickget<double>("err-translation");
  param::err_rot = options::quickget<double>("err-rotation");
  param::err_norm = options::quickget<double>("err-normal");

  param::le_min_length = options::quickget<double>("le-min-length");
  param::le_min_points = options::quickget<uint32_t>("le-min-points");
  param::le_merge_threshold = options::quickget<double>("le-merge-threshold");
  param::le_converge_threshold = options::quickget<double>("le-converge-threshold");
  param::le_tmax = options::quickget<double>("le-tmax");
  param::le_cluster_threshold = options::quickget<double>("le-cluster-threshold");

  param::da_ml_threshold = options::quickget<double>("da-threshold");
  param::spurious_discard_count = options::quickget<uint32_t>("spurious-discard-count");

  param::rectilinearity_delta = deg2rad(options::quickget<double>("delta"));

  param::gui = options::quickget<bool>("gui");

  std::string ff = options::quickget<std::string>("features-file");
  param::known_features = (ff.length() > 0);

  reset_analysis_logs();

  return inpidx;
}
