/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _SLAM_HPP
#define _SLAM_HPP

// global header file

#include <stdint.h> // for uint32_t, etc.
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <limits>
#include <vector>
#include <list>
#include <map>
#include <vl/VLd.h>
#include "fast_landmark_map.hpp"


//////////////////////////////////////////////////////////////////////
// parameters
//////////////////////////////////////////////////////////////////////

struct pose_t; // defined below

typedef enum {
  FS1,
  FS2,
  ROUGH,
  BLOCK,
  CONSTRAINED
} slam_algorithm;

typedef enum {
  RANDOM,
  RESIDUAL,
  GENERALIZED
} resample_algorithm;

namespace param
{
  // filters
  extern uint32_t N; // # particles
  extern uint32_t m; // scans per multiscan
  extern slam_algorithm slam_alg;
  extern resample_algorithm resample_alg;
  extern double resample_neff_percent;
  extern double resample_alpha; // only used for generalized resampling
  extern uint32_t lag_time; // for fixed-lag filters

  // model
  extern pose_t origin;
  extern double scan_min, scan_max, scan_res;
  extern uint32_t sparse_points;
  extern double max_range;
  extern double err_range, err_bearing;
  extern double err_trans, err_rot, err_norm;

  // feature extraction
  extern double le_min_length;
  extern uint32_t le_min_points;
  extern double le_merge_threshold;
  extern double le_converge_threshold;
  extern double le_tmax;
  extern double le_cluster_threshold;

  // data association
  extern double da_ml_threshold;
  extern uint32_t spurious_discard_count;

  // constraints
  extern double rectilinearity_delta;

  // misc
  extern bool gui;
  extern bool known_features;
  extern FILE *sample_dist_log;
  extern FILE *full_sample_dist_log;
  extern FILE *neff_log;
  extern FILE *performance_log;
};



typedef Mat2d cov2_t;
typedef Mat3d cov3_t;

//////////////////////////////////////////////////////////////////////
// simple utilities
//////////////////////////////////////////////////////////////////////

// limits
inline double max_double() { return std::numeric_limits<double>::max(); }
inline uint32_t max_uint32() { return std::numeric_limits<uint32_t>::max(); }
inline double epsilon() { return std::numeric_limits<double>::epsilon(); }

// used in initializing the motion model to guarantee that the
// distribution that comes out of the model is positive definite even
// if there's no movement
static Mat3d cov_epsilon = Mat3d(1e-20,0,0, 0,1e-20,0, 0,0,1e-20);
static Mat2d cov_infinity = Mat2d(max_double(),0, 0,max_double());

inline double deg2rad(double a) { return a * M_PI / 180; }
inline double rad2deg(double a) { return a * 180 / M_PI; }

//inline double clamp_angle(double a) { return atan2(sin(a), cos(a)); }
inline double clamp_angle(double a)
{
  while(a < -M_PI)
    a += 2 * M_PI;
  while(a >= M_PI)
    a -= 2 * M_PI;
  return a;
}

inline double angular_distance(double a1, double a2)
{
  double a = clamp_angle(a1-a2);
  if(a > M_PI) a -= 2 * M_PI;
  return a;
}

// in VL
//inline double sign(double d) { if(d < 0) return -1; else return 1; }

// in types.cpp
Mat2d mat2_from_matd(const Matd &in);
Vec2d vec2_from_vecd(const Vecd &in);
Mat3d mat3_from_matd(const Matd &in);
Vec3d vec3_from_vecd(const Vecd &in);

// in types.cpp
FILE * open_file(const char *prefix, const char *extension, const char *mode = "w");
void reset_analysis_logs();
void close_analysis_logs();

// in gaussian.cpp
struct pose_t;
double gaussian(double mean, double stddev);
pose_t gaussian(const pose_t &mean, const cov3_t &cov);

// in gaussian_cdf.cpp
double gsl_cdf_gaussian_P(const double x, const double sigma);
double gsl_cdf_gaussian_Q(const double x, const double sigma);


//////////////////////////////////////////////////////////////////////
// core types
//////////////////////////////////////////////////////////////////////

struct pose_t : public Vec3d
{
  pose_t() : Vec3d(0,0,0) {}
  pose_t(double x, double y, double t) : Vec3d(x,y,t) {}
  pose_t(const pose_t &p) : Vec3d(p) {}
  pose_t(const Vec3d &v) : Vec3d(v) {}
  double & x() { return (*this)[0]; }
  const double & x() const { return (*this)[0]; }
  double & y() { return (*this)[1]; }
  const double & y() const { return (*this)[1]; }
  double & t() { return (*this)[2]; }
  const double & t() const { return (*this)[2]; }
};

void merge(pose_t &p1, cov3_t &c1, const pose_t &p2, const cov3_t &c2);

struct pose_history_t
{
public:
  // these details are public for count_unique_particles_at_each_timestep
  struct node : public pose_t
  {
    node(const pose_t &p, node *n) : pose_t(p), next(n), count(1) {}
    node *next;     // linked list pointer
    uint32_t count; // reference count
  };
  node *head;
  void destroy();
  uint32_t length;

public:
  pose_history_t() : head(0), length(0) {}
  pose_history_t(const pose_history_t &h) : head(0) { *this = h; }
  ~pose_history_t() { destroy(); }
  void clear() { destroy(); }
  void push_back(const pose_t &p) { head = new node(p, head); ++length; }
  uint32_t size() const { return length; }
  pose_history_t & operator=(const pose_history_t &h);
  template <class List> void fixed_lag_modify(const List &lag_poses);
  void render(uint32_t how_far_back = max_uint32()) const;
  void write_to_file(const char *prefix) const;
  void read_from_file(const char *file);
  void get_stl_list(std::list<pose_t> &out) const;
  const pose_t & back() const { return *head; }
};

// change the last few poses to match lag_poses (which should be in
// reverse order of time); templated so List can be either std::list
// or std::vector
template <class List>
void pose_history_t::fixed_lag_modify(const List &lag_poses)
{
  typename List::const_iterator pi = lag_poses.begin();
  node *n = head, *p = 0;
  for(; pi != lag_poses.end(); ++pi, p = n, n = n->next) {
    if(n->count == 1) {
      n->x() = pi->x(); n->y() = pi->y(); n->t() = pi->t();
      continue;
    }
    n->count--;
    n = new node(*pi, n->next);
    if(n->next) n->next->count++;
    if(p == 0) head = n;
    else p->next = n;
  }
}



// covariance matrices

class line_t; // below

struct point_t : public Vec2d
{
  point_t() : Vec2d() {}
  point_t(double x, double y) : Vec2d(x,y) {}
  point_t(const point_t &p) : Vec2d(p) {}
  point_t(const Vec2d &v) : Vec2d(v) {}
  point_t(const pose_t &p) : Vec2d(p.x(),p.y()) {}
  double & x() { return (*this)[0]; }
  const double & x() const { return (*this)[0]; }
  double & y() { return (*this)[1]; }
  const double & y() const { return (*this)[1]; }
};

template <class F>
inline double squared_distance(const point_t &p, const F &f)
{
  return sqrlen(p-f);
}
template <class F>
inline double distance(const point_t &p, const F &f)
{
  return len(p-f);
}
point_t transform(const point_t &p, const pose_t &delta);

struct line_t : public Vec2d
{
  line_t() : Vec2d() {}
  line_t(double r, double t) : Vec2d(r,t) {}
  line_t(const line_t &l) : Vec2d(l) {}
  line_t(const Vec2d &v) : Vec2d(v) {}
  double & r() { return (*this)[0]; }
  const double & r() const { return (*this)[0]; }
  double & t() { return (*this)[1]; }
  const double & t() const { return (*this)[1]; }
};
line_t transform(const line_t &l, const pose_t &delta);

// gets point on line from dist along line
inline point_t u_to_point(const line_t &l, double u)
{
  point_t ret;
  ret.x() = l.r()*cos(l.t()) + u*cos(l.t()+M_PI/2);
  ret.y() = l.r()*sin(l.t()) + u*sin(l.t()+M_PI/2);
  return ret;
}

// projects p onto the line and returns dist along line
inline double project_onto(const point_t &p, const line_t &l)
{
  return (p.x() - l.r()*cos(l.t())) * cos(l.t()+M_PI/2) +
    (p.y() - l.r()*sin(l.t())) * sin(l.t()+M_PI/2);
}

inline double squared_distance(const point_t &p, const line_t &f)
{
  double u = project_onto(p, f);
  point_t pt(u_to_point(f, u));
  return squared_distance(p, pt);
}


// sensor scan

typedef std::vector<double> scan_t;
struct cartesian_scan_t : public std::vector<point_t>
{
  pose_t frame;
  cartesian_scan_t() {}
  cartesian_scan_t(const scan_t &in, const pose_t &_frame = pose_t());
  void render() const;
  void radial_sort(const pose_t &p); // sort by angle wrt p
  void transform(const pose_t &delta);
};
typedef std::vector<cartesian_scan_t> multiscan_t;

void agglomerate_multiscan(const multiscan_t &ms, cartesian_scan_t &out);


// features/landmarks

struct line_feature : public line_t
{
  cov2_t cov;

  line_feature() {}
  line_feature(const line_feature &f) : line_t(f), cov(f.cov) {}
  line_feature(const line_t &f) : line_t(f) {}
  void render() const;
};

double mahalanobis_distance(const line_feature &f1, const line_feature &f2);
line_feature merge(const line_feature &f1, const line_feature &f2);
line_feature transform(const line_feature &f, const pose_t &delta);


template <class Feature>
double equality_likelihood_from_mahalanobis
  (const Feature &f1, const Feature &f2, double md)
{
  double lh = exp(-0.5*md*md) / (::sqrt(det(f1.cov+f2.cov)) * 2*M_PI);
  if(isnan(lh) || lh == 0)
    return epsilon();
  return lh;
}

template <class Feature>
double equality_likelihood(const Feature &f1, const Feature &f2)
{
  return equality_likelihood_from_mahalanobis(f1, f2, mahalanobis_distance(f1, f2));
}


struct extent_feature : public line_feature
{
  double umin, umax;

  extent_feature() {}
  extent_feature(const extent_feature &f) : line_feature(f), umin(f.umin), umax(f.umax) {}
  extent_feature(const line_feature &f) : line_feature(f) {}
  void render(bool render_uncertainty = false) const;
  void write_to_file(FILE *file) const;
  void read_from_file(FILE *file);
};

double squared_length(const extent_feature &f);
double length(const extent_feature &f);
double squared_distance(const point_t &p, const extent_feature &f);
bool overlap(const extent_feature &f1, const extent_feature &f2);
extent_feature merge(const extent_feature &f1, const extent_feature &f2);
extent_feature transform(const extent_feature &f, const pose_t &delta);
void feature_to_particle_frame
  (const extent_feature &in, const pose_t &in_frame,
   extent_feature &out, const pose_t &out_frame);


struct point_feature : public point_t
{
  cov2_t cov;

  point_feature() {}
  point_feature(const point_feature &f) : point_t(f), cov(f.cov) {}
  point_feature(const point_t &f) : point_t(f) {}
  void render(bool render_uncertainty = false) const;
  void write_to_file(FILE *file) const;
  void read_from_file(FILE *file);
};

double mahalanobis_distance(const point_feature &f1, const point_feature &f2);
inline bool overlap(const point_feature &f1, const point_feature &f2) { return true; }
point_feature merge(const point_feature &f1, const point_feature &f2);
point_feature transform(const point_feature &f, const pose_t &delta);
void feature_to_particle_frame
  (const point_feature &in, const pose_t &in_frame,
   point_feature &out, const pose_t &out_frame);


//////////////////////////////////////////////////////////////////////
// global typedefs: now we've got all the basic types, mainly the
// feature types.  here's where to change things if you want to, say,
// switch between point and extent features
//////////////////////////////////////////////////////////////////////

typedef point_feature feature_type;
//typedef extent_feature feature_type;
//#define _LINE_FEATURES // FIXME remove
typedef map_t<feature_type> map_type;
// make sure you define constraint_type appropriately in rbpf.hpp!


//////////////////////////////////////////////////////////////////////
// feature extraction
//////////////////////////////////////////////////////////////////////

// extract features in ABSOLUTE coordinate frame --- thus, the
// cartesian points should be wrt the world frame (alternatively, they
// could be relative to the robot frame, and then the features could
// be transformed into the world frame)
void extract_features(const cartesian_scan_t &points, std::vector<extent_feature> &out);

// a dummy for now
void extract_features(const cartesian_scan_t &points, std::vector<point_feature> &out);


//////////////////////////////////////////////////////////////////////
// data association
//////////////////////////////////////////////////////////////////////

// computes the best data association of f among landmarks in
// in_range; sets assoc and mahalanobis_dist (to the MD between f and
// assoc)
void data_association
  (const feature_type &f,
   const std::vector<map_type::iterator> &in_range,
   map_type::iterator &assoc, double &mahalanobis_dist);


//////////////////////////////////////////////////////////////////////
// other basic slam stuff
//////////////////////////////////////////////////////////////////////

// compute basic proposal distribution from motion model

// odometry based motion model
void motion_model
  (const pose_t &delta, const pose_t &ref_frame, const pose_t &particle_frame,
   pose_t &mu, cov3_t &cov);

// detect landmarks we should have seen but didn't, and remove them;
// some potential approaches: brute-force raytracing to test all
// in-range landmarks; stochastic testing of a few randomly-selected
// in-range landmarks
void remove_spurious_landmarks
  (const pose_t &p, map_type &m,
   std::vector<map_type::iterator> &in_range,
   const cartesian_scan_t &points,
   std::map<map_type::iterator, int32_t> &obs_count);


#endif // _SLAM_HPP
